defmodule Archeometer.Analysis.Apps.Xref do
  @moduledoc """
  Functions for generating a dependency graph from applicatios within an umbrella application.

  Accepted output formats are "dot" (graphviz), "png" and "mermaid".
  """

  alias Archeometer.Graphs.Graphviz
  alias Archeometer.Graphs.Mermaid

  @supported_formats ["png", "dot", "mermaid"]

  @doc """
  Creates a dependency graph between the applications of the current project.

  ## Parameters

  - `format` can be one of "dot" (graphviz), "png", or "mermaid".

  ## Returns

  - The binary representing the graph, if the operation was completed successfully.
  - `{:error, reason}` if not.

  """
  def gen_graph(format) when format in @supported_formats do
    do_gen_graph(format, Mix.Project.umbrella?())
  end

  def gen_graph(_format) do
    {:error, :unsupported_format}
  end

  defp do_gen_graph(_format, false) do
    {:error, :not_umbrella_project}
  end

  defp do_gen_graph(format, true) do
    Mix.Project.apps_paths()
    |> Enum.map(&xrefs/1)
    |> Enum.into(%{})
    |> render(format)
  end

  defp xrefs({app_id, app_rel_path}) do
    app_path = Path.join(File.cwd!(), app_rel_path)
    xrefs = Mix.Project.in_project(app_id, app_path, &refs/1)
    {app_id, xrefs}
  end

  defp refs(_module) do
    Mix.Project.config()
    |> Keyword.get(:deps)
    |> Enum.filter(&in_umbrella?(&1))
    |> Enum.map(fn {dep_id, _} -> dep_id end)
  end

  defp in_umbrella?(dep_config) do
    case dep_config do
      {_, _, _} ->
        false

      {_dep_id, cfg} ->
        Keyword.keyword?(cfg) and Keyword.get(cfg, :in_umbrella, false)
    end
  end

  defp render(refs, "dot") do
    Graphviz.render_dot(refs)
  end

  defp render(refs, "png") do
    Graphviz.render_png(refs)
  end

  defp render(refs, "mermaid") do
    Mermaid.render(refs)
  end
end
